//
// Created by brett on 25/10/23.
//

#ifndef INSANE_DNS_IP_H
#define INSANE_DNS_IP_H

#include <blt/std/string.h>
#include <blt/compatibility.h>
#include <blt/std/assert.h>

#include <asio.hpp>

struct IPAddress
{
    unsigned char octets[4]{};
    std::string asString;
    
    BLT_CPP20_CONSTEXPR IPAddress(std::string str)
    {
        auto data = blt::string::split(str, '.');
        BLT_ASSERT(data.size() == 4);
        for (size_t i = 0; i < data.size(); i++)
            octets[i] = static_cast<unsigned char>(std::stoul(data[i]));
        asString = std::move(str);
    }
    
    BLT_CPP20_CONSTEXPR IPAddress(const unsigned char oct[4])
    {
        // ugly sin
        for (int i = 0; i < 4; i++)
            octets[i] = oct[i];
    }
};

struct bind_address
{
    unsigned short int port;
    bool use_v4 = true;
    bool use_v6 = true;
    // if these are empty it will use ASIO default
    std::string IP_V4;
    std::string IP_V6;
};

// using a template instead of a macro. Are you proud of me dad?
template<typename RET, typename V4, typename V6>
inline RET BLT_INTERNAL_ENDPOINT(const bind_address& address, V4 v4, V6 v6)
{
    // this is an ugly mess. Ignore it.
    BLT_ASSERT(address.use_v4 || address.use_v6);
    if (address.use_v6)
    {
        if (address.IP_V6.empty())
            return {v6, address.port};
        else
            return {asio::ip::address::from_string(address.IP_V6), address.port};
    }
    if (address.use_v4)
    {
        if (address.IP_V4.empty())
            return {v4, address.port};
        else
            return {asio::ip::address::from_string(address.IP_V4), address.port};
    }
    // suppress compiler warnings. Also, could do std::exit() but this is more fun :3
    return {asio::ip::address::from_string("Unreachable Code"), 42069};
}

/**
 * Convert a bind_address into a ASIO UDP endpoint
 * @param address address to convert
 * @return ASIO UDP endpoint
 */
inline asio::ip::udp::endpoint toUDPEndpoint(const bind_address& address)
{
    auto v4 = asio::ip::udp::v4();
    auto v6 = asio::ip::udp::v6();
    return BLT_INTERNAL_ENDPOINT<asio::ip::udp::endpoint>(address, v4, v6);
}

/**
 * Convert a bind_address into a ASIO TCP endpoint
 * @param address address to convert
 * @return ASIO TCP endpoint
 */
inline asio::ip::tcp::endpoint toTCPEndpoint(const bind_address& address)
{
    auto v4 = asio::ip::tcp::v4();
    auto v6 = asio::ip::tcp::v6();
    return BLT_INTERNAL_ENDPOINT<asio::ip::tcp::endpoint>(address, v4, v6);
}

struct request_info
{
    size_t number_of_bytes;
    size_t number_of_answers;
};

#endif //INSANE_DNS_IP_H
